import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.openstreetmap.josm.data.coor.EastNorth;
import org.openstreetmap.josm.data.osm.Node;
import org.openstreetmap.josm.data.osm.OsmPrimitive;
import org.openstreetmap.josm.tools.Pair;


/*
 * This finds nodes close to other nodes.  It does it by keeping
 * an index of all nodes in two indexes: on by X coordinate and
 * one by Y.
 */
public class NodeSearchCache
{
    private SortedMap<Double, Object> by_x;
    //private SortedMap<Double, Node> by_y;
    // If we could find a nice Collection data type that would
    // quickly return a sorted list of keys, we would not
    // need these.
    ArrayList<Double> x_s;
    //ArrayList<Double> y_s;
    public NodeSearchCache()
    {
        by_x = new TreeMap<Double, Object>();
        //by_y = new TreeMap<Double, Node>();
    }
    public void add(OsmPrimitive n)
    {
        add((Node)n);
    }
    public void add(Node n)
    {
        by_x.put(new Double(n.x), new Pair<Double, Node>(n.y, n)); 
        //by_y.put(new Double(en.north()), n);
        x_s = null;
        //y_s = null;
    }
    private Double pair_to_double(Pair p)
    {
        return (Double)p.a;
    }
    private void store(Object o, SortedMap<Double, Object> map, Double d)
    {
        if (o instanceof Pair) {
            Pair p = (Pair)o;
            Double key = pair_to_double(p);
            Object val = p.b;
            map.put(key, val);
            return;
        }
        map.put(d, o);
    }
    SortedMap<Double, Object> getNear(SortedMap<Double, Object> map,
                             ArrayList<Double> list, double value, double distance)
    {
        // Do a binary search to find a good spot in the list
        int index = Collections.binarySearch(list, value);
        SortedMap<Double, Object> res = new TreeMap<Double, Object>();
        
        // The negative index says that "the 'value' was not found
        // but it would have been inserted hete"
        if (index < 0)
            index = -index + 1;
        // binarySearch() cal also return a value at list.size(),
        // so make sure to make this into a valid index
        if (index >= list.size())
            index = list.size()-1;
        int low = index;

        // Now iterate out from the binary search spot to find
        // other points that might be close.
        while (list.get(low).doubleValue() >= (value - distance)) {
            Object o = map.get(list.get(low));
            store(o, res, list.get(low));
            low--;
            if (low < 0)
                break;
        }
        int high = index;
        while (list.get(high).doubleValue() <= (value + distance)) {
            Object o = map.get(list.get(high));
            store(o, res, list.get(high));
            high++;
            if (high >= list.size())
                break;
        }
        return res;
    }
    static int nr = 0;
    double total_count = 0;
    double total_count_r1 = 0;
    static int nr_r1 = 0;
    double total_count_r2 = 0;
    static int nr_r2 = 0;
    public Collection<OsmPrimitive> getNear(double x, double y, double _distance)
    {
        //nr++;
        // This is heavily optimized right now to have all the
        // nodes added, then have searches done.  If they're mixed
        // then this array creation might start to get more costly
        if (x_s == null) {
            Set <Double> xKeys = by_x.keySet();
            x_s = new ArrayList<Double>(xKeys);
        }
        //if (y_s == null) {
        //    Set <Double> yKeys = by_y.keySet();
        //    y_s = new ArrayList<Double>(yKeys);
        //}
        //total_count += y_s.size();
        SortedMap<Double, Object> xset = getNear(by_x, x_s, x, _distance);
        //nr_r1++;
        //total_count_r1 += xset.size();
        if (xset.size() == 0)
            return new ArrayList<OsmPrimitive>();
        ArrayList<Double> tmp_y_s = new ArrayList<Double>(xset.keySet());
        SortedMap<Double, Object> yset = getNear(xset, tmp_y_s, y, _distance);
        /*
        nr_r2++;
        total_count_r2 += yset.size();
        if (nr % 100 == 0) {
            System.out.println("count: "+nr+" avg starting size: " + (total_count/nr));
            System.out.println("r1 avg size: " + (total_count_r1/nr_r1));
            System.out.println("r2 avg size: " + (total_count_r2/nr_r2));
        }*/
        ArrayList<OsmPrimitive> ret = new ArrayList<OsmPrimitive>();
        for (Object o : yset.values())
            ret.add((Node)o);
        return ret;
    }
}
